/* pbm2oled - generate C source code fragments from "portable bitmap" image data

  This program generates a C source code fragment that describes the contents of
  a monochrome "portable bitmap" ("PBM") image file, in a format compatable with
  the (ATTiny85 Arduino) DigisparkOLED bitmap() method, for presentation on a
  GM009605v4 128x64 IIC OLED display.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

/* NETPBM library includes */
#include <netpbm/pam.h>

void Pbm2Oled(char *, int, int, char *);
void writeComment(int, int, bit **, int);
void writeBitmap(int, int, bit **, int, int, char *);

int main(int argc, char *argv[])
{
  int	opt,
	ok = 1,		/* 0 -> errors, 1 -> no errors yet */
	invert = 0,	/* display PBM_BLACK as white, PBM_WHITE as black */
	comment = 0;	/* generate representation of image as C comment */
  char	*defn_name = NULL,
	*pbm_file = NULL;

  extern int 	optind,
		opterr,
		optopt;
  extern char	*optarg;

  pm_init(argv[0],0);	/* required NETPBM initialization */

  opterr = 0;	/* suppress getopt error msgs */
  while ((opt = getopt(argc,argv,"cin:")) != -1)
  {
    switch (opt)
    {
      case 'c':	/* generate image in C comment */
	comment = 1;
	break;

      case 'i':	/* invert Black and White */
	invert = 1;
	break;

      case 'n':	/* gives output defn a name */
	if (defn_name)
	{
	  fprintf(stderr,"Error: Output data cannot have multiple names - \"%s\" rejected\n",optarg);
	  ok = 0;
	}
	else defn_name = strdup(optarg);
	break;

      default: /* unidentified argument */
	switch (optopt)
	{
	  case 'n':
	    fprintf(stderr,"Error: -n requires additional information\n");
	    break;

	  default:
	    fprintf(stderr,"Error: Unknown '%c' flag\n",optopt);
	    break;
	}
	ok = 0;
	break;
    }
  }

  if (ok)
    if (optind < argc)
      if (optind != argc-1)
      {
	/* too many non-flag arguments */
	fputs("Error: too many non-flag arguments\n",stderr);
	ok = 0;
      }
      else pbm_file = argv[optind];

  if (ok)
  {
    if (pbm_file == NULL)	pbm_file = "-";
    if (defn_name == NULL)	defn_name = "image";
    Pbm2Oled(pbm_file,invert,comment,defn_name);
  }
  else fprintf(stderr,"Usage: %s [-c] [-i] [-n name] [pbmfile]\n",argv[0]);

  return ok?EXIT_SUCCESS:EXIT_FAILURE;
}

/*
** Pbm2Oled loads the designated PBM file into memory, and emits C code to
** represent the loaded image in the format required by the DigisparkOLED
** library's bitmap() method.
*/
void Pbm2Oled(char *fname, int invert, int comment, char *target)
{
  FILE	*pbm;

  if (pbm = pm_openr(fname))
  {
    bit **image;
    int pbm_rows, pbm_cols;

    image = pbm_readpbm(pbm,&pbm_cols,&pbm_rows);
    pm_close(pbm);

    if (comment) writeComment(pbm_rows, pbm_cols, image, invert);
    writeBitmap(pbm_rows, pbm_cols, image, invert, comment, target);

    pbm_freearray(image,pbm_rows);
  }
}

/*
** writeComment() writes to stdout a C comment containing a graphic
** representation of the bitmapped image.
*/
void writeComment(int pbm_rows, int pbm_cols, bit **image, int invert)
{
  int	row;
  char	*hilite;

  if (invert)
    hilite= "O.";
  else
    hilite= ".O";

  puts("  /*");
  for (row = 0; row < pbm_rows; ++row)
  {
    int col;

    putchar(' '); putchar(' '); putchar(' '); putchar(' ');
    for (col = 0; col < pbm_cols; ++col)
    {
      if (image[row][col] == PBM_WHITE)
	putchar(hilite[0]);
      else
	putchar(hilite[1]);
    }
    putchar('\n');
  }
  puts("  */");
}

/*
** writeBitmap() writes to stdout the C code needed to store a bitmapped image
** for future display using the DigisparkOLED library bitmap() method.
**
** writeBitmap()
**   Accepts a PBM image size in rows and columns, a PBM image, and a set of processing flags
**   Processes the given PBM image 8 rows at a time, outputting 8-bit image chunks where
**     the LSB of the chunk represents the pixel at row+0, column, and the MSB of the chunk
**     represents the pixel at row+7,column. The function processes the image left-to-right,
**     top to bottom.
**   Returns nothing.
**
** Note writeBitmap() processes the PBM 8 rows at a time, left to right, top to bottom
**	Each byte of the output bitmap represents a single column of 8 pixels, with the
**	high-order bit representing the bit closest to the top.
**
**	Thus, a byte value of 0xA6 represents 8 pixels in a column, which would look like
**
**	   ___________bit positions in bitmap byte (76543210)
**	  /     ______bitmap byte in binary
**       /     / 
**	7  +- 1  #
**	6  A  0  -
**	5  :  1  #
**	4  +- 0  -
**	3  +- 0  -
**	2  6  1  #
**	1  :  1  #
**	0  +- 0  -
**          \     \__displayed on OLED
**	     \_______bitmap byte in hex
*/
void writeBitmap(int pbm_rows, int pbm_cols, bit **image, int invert, int comment, char *target)
{
  int	rows,
	cols,
	size,
	row;
  bit	BorW;

  if (invert)
    BorW= PBM_WHITE;
  else
    BorW= PBM_BLACK;

  rows = (7 + pbm_rows) / 8;
  cols = pbm_cols;
  size = rows * cols;

  putchar('\n');
  printf("  const struct { unsigned int rows, cols; unsigned char map[%d]; } %s PROGMEM =\n",size,target);
  printf("  {\n    %d,%d,\n    {\n",rows,cols);

  for (row = 0; row < pbm_rows; row+= 8)	/* each output octet represents 1 column of 8 image rows */
  {
    int col;

    printf("      ");
    for (col = 0; col < pbm_cols; ++col)
    {
      unsigned char map = 0;
      int subrow;

      for (subrow = 7; subrow > -1; --subrow)
      {
        map <<= 1;

        if (row+subrow < pbm_rows)
	{
	   if (image[row+subrow][col] == BorW) map|= 1;
	}
      }
      printf("0x%2.2x, ",map);
    }
    puts("");
  }
  puts("    }\n  };");
  printf("  /* bitmap(hoz,vrt,hoz+%s.cols,vrt+%s.rows,%s.map); */\n",target,target,target);
}

/*  Copyright (C) 2020 Lew Pitcher <lew.pitcher@digitalfreehold.ca>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/
